#ifndef DCM_VERSION_H
#define DCM_VERSION_H

/* The project version.
 */
#define DCM_VERSION            "1.3.0"
#define DCM_SUFFIXED_VERSION   "1.3.0"
#define DCM_VERSION_MAJOR      (1)
#define DCM_VERSION_MINOR      (3)
#define DCM_VERSION_MICRO      (0)

/* The ABI version, as used for library versioning.
 */
#define DCM_ABI_VERSION_MAJOR  (1)
#define DCM_ABI_VERSION_MINOR  (3)
#define DCM_ABI_VERSION_PATCH  (0)

/* Whether libdicom is built statically.
 */
#define DCM_STATIC             (0)

#endif /*DCM_VERSION_H*/


